"""This module contains a class Message, which is used to represent a chat
message. A Message object can format itself for transfer. This module also
exports a function unformat, which does the inverse operation."""

# David Fifield, 2007

def unescape(s):
	"""Interpret %XX hexadecimal sequences as raw characters."""
	output = []
	i = 0
	while i < len(s):
		c = s[i]
		i += 1
		if c == "%":
			try:
				h = s[i] + s[i + 1]
			except IndexError:
				raise FormatError, "End of string found while expecting escaped character."
			try:
				c = chr(int(h, 16))
			except ValueError:
				raise FormatError, "Expected hexadecimal digits, got \"%s\"." % h
			i += 2
		output.append(c)
	return "".join(output)

def escape_char(c):
	"""Return the %XX hexdecimal escape of a character."""
	return "%%%02X" % ord(c)

def escape(s, chars = ""):
	"""Escape each of the characters in chars in s."""
	output = []
	for c in s:
		if c in chars or c == "%":
			output.append(escape_char(c))
		else:
			output.append(c)
	return "".join(output)

def escape_body(s):
	"""Escape the characters necessary for a body."""
	s = escape(s)
	lines = []
	for line in s.split("\n"):
		if line == ".":
			line = escape_char(".")
		lines.append(line)
	return "\n".join(lines)

def escape_header_field_name(s):
	"""Escape the characters necessary for a header field name."""
	return escape(s, "\n:")

def escape_header_field_value(s):
	"""Escape the characters necessary for a header field value."""
	s = escape(s, "\n")
	# Preserve leading whitespace by escaping the first character.
	if len(s) > 0 and s[0].isspace():
		s = escape_char(s[0]) + s[1:]
	return s

def unformat(formatted):
	"""Return a Message object from a formatted string representation.
	
	Raise a FormatError exception on any format error.
	"""

	# Make sure the string ends with the proper terminator, then strip it.
	if not formatted.endswith(Message.TERMINATOR):
		raise FormatError, "Formatted message does not have a terminator."
	formatted = formatted[:-3]

	m = Message()

	# Separate the formatted message into a header and a body.
	try:
		escaped_header, escaped_body = formatted.split(Message.SEPARATOR, 1)
	except ValueError:
		raise FormatError, "Formatted message does not have a separator."

	# Break each header field into a name and a value and store them in the
	# message.
	if escaped_header != "":
		escaped_header_fields = escaped_header.split("\n")
		for escaped_header_field in escaped_header_fields:
			try:
				escaped_name, escaped_value = escaped_header_field.split(":")
			except ValueError:
				raise FormatError, "Header field in formatted message does not contain \":\"."
			name = unescape(escaped_name)
			value = unescape(escaped_value.lstrip())
			m.header[name] = value

	# Unescape the body and store it in the message.
	body = unescape(escaped_body)
	m.body = body

	return m

class FormatError(Exception):
	"""An exception that is raised when attempting to unformat an improperly
	formatted message."""
	pass

class StringDict(dict):
	"""A dictionary class that is constrained to holding strings."""
	def __setitem__(self, key, value):
		dict.__setitem__(self, str(key), str(value))

class Message:
	"""A message.

	Each message consists of a header and a body. The header is accessible as a
	dictionary. Each of the header field names, header field values, and the
	body are converted to plain strings before being stored.
	
	This class can format a message into a byte string for transfer over a
	network. (The module-level function unformat does the reverse operation.)
	The formatting is as follows:
	
	The header is a sequence of "Name: Value" pairs separated by single
	newlines. The header is separated from the body by a pair of newlines
	("\n\n"). The body is terminated by a "\n.\n" string. Characters may be
	escaped using %XX hexadecimal sequences. This is an example of a message:

		ID: user
	
		Hi. This is my message.
		.
	"""

	# The string that separates the header of a formatted message from its
	# body. It is guaranteed to be present at least once in each formatted
	# message. The first one separates the header from the body.
	SEPARATOR = "\n\n"

	# The string that separates one formatted message from another in the data
	# stream. It is guaranteed to be present exactly once at the end of each
	# formatted message.
	TERMINATOR = "\n.\n"

	def __init__(self, header = {}, body = ""):
		"""Initialize a message."""
		self.header = StringDict()
		self.header.update(header)
		self.__setbody(body)

	def __cmp__(self, other):
		if isinstance(other, Message):
			return cmp(self.header, other.header) or cmp(self.body, other.body)
		else:
			return cmp(self, other)

	def __setbody(self, body):
		self.body = str(body)

	def format(self):
		"""Return a formatted string representing this message."""
		header_string = "\n".join("%s: %s" \
			% (escape_header_field_name(n), escape_header_field_value(v)) \
			for n, v in self.header.items())
		body_string = escape_body(self.body)
		return header_string + Message.SEPARATOR \
			+ body_string + Message.TERMINATOR
